globals [
 layers1
 layers2

 input-node-list1
 output-node-list1
 hidden-nodes1

 input-node-list2
 output-node-list2
 hidden-nodes2

 layer-counts

 meaning-language
 message-language
 language-model

 mapping1-training-n
 mapping2-training-n
 learning-rate

]

turtles-own [
 activation
 bias

 target
 err
]

links-own [
 weight
]

to test-setup

  setup-brain ["o0" "o1" "o2" "r0"] num-middle-nodes 2 ["w0" "w1" "w2" "w3"] 100 100 .5
  setup-language-model

end

to test-update

  update-mappings language-model
  ask turtles [recolor]
  ask links   [update-link-look]

  let i one-of [0]
  let message item 1 (item i language-model)
  let meaning item 0 (item i language-model)

  print (apply-bools2 apply-bools1 meaning) = (meaning)

end

to setup-language-model

  ;let meanings map [replace-item ? [false false false false] true] shuffle [0 1 2 3]
  ;let messages map [replace-item ? [false false false false] true] shuffle [0 1 2 3]
  ;set language-model (map [(list ?1 ?2)] meanings messages)
  set language-model [[[false true true false] [false true false false]]]

end

to setup-brain [meaning-input middle-layer middle-layers message-input train-input1 train-input2 learning-rate-input]

  ;; creates the brain of the turtle
  ;; note that this is repurposed from Sheeps with Brains (thanks guys!)

  ca
  set-default-shape turtles "circle"

  ;; save the language model
  set meaning-language meaning-input
  set message-language message-input

  ;; save training settings
  set mapping1-training-n train-input1
  set mapping2-training-n train-input2
  set learning-rate learning-rate-input

  let middle-layer-tmp map [middle-layer] n-values middle-count [?]
  set layer-counts  flatten-list (list (length meaning-language) middle-layer-tmp (length message-language))
  set layers1 []
  set layers2 []
  let max-layer-size max ( layer-counts )

  ;; resize world
  let max-x 30
  let max-y max (list ((max-layer-size * 2) + 5) 15)
  resize-world 0 max-x (- max-y) max-y

  setup-network-1
  setup-network-2
  randomize-weights

  reset-ticks

end

to setup-network-1
   ;; create meaning to message network
  ;; since the layers came in the right order, then
  ;; they are created in the right order, as well
  foreach layer-counts [
    let layer-size ? ;; sets the
    let layer-num length layers1 ;; figures out what layer the code is on
    let layer no-turtles ;; initializes an agent set to  be added to

    ;; creates LAYER-SIZE nodes and appends them to the TURTLE-SET agentset
    crt layer-size [ set layer (turtle-set layer self) ]

    ;; figures out the x-coordinate for the nodes in this layer
    ;; Since all the nodes are placed in a column, they can share
    ;; an x-coordinate.
    let x ( (layer-num + 0.5) / (middle-count + 2) ) * max-pxcor

    ;; figures out where the nodes are to be placed along the y-axis, then
    ;; does so
    let i 0

    ;; sort converts the elements of the layer, an agent-set, into a list
    ;; to be iterated over.
    ;; This is done instead of using an ask in order to access the node's
    ;; index, so they can be placed accordingly.
    foreach sort layer [
      ask ? [
        let y ((2 * (0.5 - (i + 0.25) / layer-size)) * max-pycor / 2) + (1 / 2) * max-pycor
        setxy x y
        set color green
      ]
      set i i + 1
    ]

    set layers1 lput layer layers1
  ]

  ;; iterates over the middle layers of the network
  (foreach (but-first layers1) (but-last layers1) [
    ask ?1 [
      create-links-from ?2 [
        set-weight 0
      ]
    ]
  ])

  ;; figures out which are the input and
  ;; which are the output
  set input-node-list1 sort first layers1
  set output-node-list1 sort last layers1

  ;; get other agentsets, er, set
  set hidden-nodes1 last but-last layers1
end

to setup-network-2
   ;; create message to meaning network
  ;; since the layers came in the right order, then
  ;; they are created in the right order, as well

  ;; note that since this is the message to mapping network, we have to flip
  ;; the layer counts
  foreach reverse layer-counts [
    let layer-size ?
    let layer-num length layers2 ;; figures out what layer the code is on
    let layer no-turtles ;; initializes an agent set to  be added to

    ;; creates LAYER-SIZE nodes and appends them to the TURTLE-SET agentset
    crt layer-size [ set layer (turtle-set layer self) ]

    ;; figures out the x-coordinate for the nodes in this layer
    ;; Since all the nodes are placed in a column, they can share
    ;; an x-coordinate.
    let x ( (layer-num + 0.5) / (middle-count + 2) ) * max-pxcor

    ;; figures out where the nodes are to be placed along the y-axis, then
    ;; does so
    let i 0

    ;; sort converts the elements of the layer, an agent-set, into a list
    ;; to be iterated over.
    ;; This is done instead of using an ask in order to access the node's
    ;; index, so they can be placed accordingly.
    foreach sort layer [
      ask ? [
        let y ((2 * (0.5 - (i + 0.25) / layer-size)) * max-pycor / 2) - (1 / 2) * max-pycor
        setxy x y
        set color yellow
      ]
      set i i + 1
    ]

    set layers2 lput layer layers2
  ]

  ;; iterates over the middle layers of the network
  (foreach (but-first layers2) (but-last layers2) [
    ask ?1 [
      create-links-from ?2 [
        set-weight 0
      ]
    ]
  ])

  ;; figures out which are the input and
  ;; which are the output
  set input-node-list2 sort first layers2
  set output-node-list2 sort last layers2

  ;; get hidden nodes agentset
  set hidden-nodes2 last but-last layers2
end


to-report apply-bools1 [ input ]
  let result apply-reals1 map [ ifelse-value ? [1][0] ] input
  report map [? > 0.5] result
end

to-report apply-reals1 [ input ]
  set-input1 input
  propogate1
  report map [ [ activation ] of ? ] output-node-list1
end

to set-input1 [ input ]
  (foreach input-node-list1 input [
    ask ?1 [ set-activation ?2 ]
  ])
end


to propogate1
  foreach but-first layers1 [
    ask ? [ update-activation ]
  ]
end

to-report apply-bools2 [ input ]
  let result apply-reals2 map [ ifelse-value ? [1][0] ] input
  report map [? > 0.5] result
end

to-report apply-reals2 [ input ]
  set-input2 input
  propogate2
  report map [ [ activation ] of ? ] output-node-list2
end

to set-input2 [ input ]
  (foreach input-node-list2 input [
    ask ?1 [ set-activation ?2 ]
  ])
end

to propogate2
  foreach but-first layers2 [
    ask ? [ update-activation ]
  ]
end

to update-mappings [ memory ]

  foreach memory [
       ;; train the meaning to message layer
       ;; to better approximate what the interlocutor
       ;; was expecting

       let meaning item 0 ?
       let target-message item 1 ?

       update-mapping1 meaning target-message

       ;; train the message to meaning layer
       ;; we take the meaning that we started with as the
       ;; target, and the message that our own network
       ;; generates as the base
       update-mapping2 meaning

  ]


end

to update-mapping1 [meaning target-message]

  ;; convert the boolean string to activation levels
  let int-target-message map [ ifelse-value ? [1][0] ] target-message

  ;; get the input layer set properly
  let trash apply-bools1 meaning

  repeat mapping1-training-n [

    (foreach output-node-list1 int-target-message [
      ask ?1 [ set target ?2]
      ask ?1 [ set err [activation] of ?1 * (1 - [activation] of ?1) * ([target] of ?1 - [activation] of ?1)]
    ])

    ;; The hidden layer nodes are given error values adjusted appropriately for their
    ;; link weights
    foreach sort hidden-nodes1 [

      ask ? [
        set err activation * (1 - activation) * sum [ weight * [ err ] of end2 ] of my-out-links ]
    ]
    ask links with [member? end2 hidden-nodes1]  [
      set weight weight + learning-rate * [ err ] of end2 * [ activation ] of end1
    ]

  ]

end

to update-mapping2 [ target-meaning ]

  ;; convert the boolean list into a list
  ;; of activation levels
  let int-target-meaning map [ ifelse-value ? [1][0] ] target-meaning

  ;; set the input layer
  let trash apply-bools2 apply-bools1 target-meaning

  repeat mapping2-training-n [

    (foreach output-node-list2 int-target-meaning [
      ask ?1 [ set target ?2]
      ask ?1 [ set err [activation] of ?1 * (1 - [activation] of ?1) * ([target] of ?1 - [activation] of ?1)]
    ])

    ;; The hidden layer nodes are given error values adjusted appropriately for their
    ;; link weights
    foreach sort hidden-nodes2 [

      ask ? [
        set err activation * (1 - activation) * sum [ weight * [ err ] of end2 ] of my-out-links ]
    ]
    ask links with [member? end2 hidden-nodes2]  [
      set weight weight + learning-rate * [ err ] of end2 * [ activation ] of end1
    ]

  ]


end

;; general
to set-activation [ a ]
  set activation a
  recolor
end

to set-weight [ w ]
  set weight w
  update-link-look
end

to update-activation
  let incoming sum [ weight * [ activation ] of other-end ] of my-in-links
  set-activation activation-function (bias + incoming)
end

to set-bias [ b ]
  if not member? self input-node-list1 and not member? self input-node-list2 [
    set bias b
    set label precision b 1
  ]
end

to randomize-weights
  ask links [ set-weight random-normal 0 1 ]
  ask turtles [ set-bias random-normal 0 1 ]
  display
end

;; function, er, functions
to-report activation-function [ x ]
  report sigmoid x
end

to-report sigmoid [x]
  report 1 / (1 + e ^ (- x))
end

;; helper functions

to-report flatten-list [ lst ]
   ; flattens nested lists to a single list.
   if (reduce [?1 or is-list? ?2] fput false lst) [
     set lst reduce [sentence ?1 ?2] lst
     set lst flatten-list lst
   ]
   report lst
 end

;; aesthetics functions
to update-link-look
  ;; given full activation of source
  let w activation-function weight
  if w > 0.5 [
    set color red
  ]
  if w = 0.5 [
    set color grey
  ]
  if w < 0.5 [
    set color blue
  ]
  set thickness abs (w - 0.5)
end

to recolor
  ifelse activation > 0.5 [
    set color yellow
  ] [
    set color grey
  ]
  set size 0.5 + activation
end
@#$#@#$#@
GRAPHICS-WINDOW
210
10
623
704
-1
25
13.0
1
10
1
1
1
0
1
1
1
0
30
-25
25
0
0
1
ticks
30.0

BUTTON
47
165
143
198
NIL
test-setup
NIL
1
T
OBSERVER
NIL
NIL
NIL
NIL
1

BUTTON
44
113
148
146
NIL
test-update
T
1
T
OBSERVER
NIL
NIL
NIL
NIL
1

SLIDER
14
55
186
88
middle-count
middle-count
0
4
4
1
1
NIL
HORIZONTAL

SLIDER
12
11
186
44
num-middle-nodes
num-middle-nodes
1
10
10
1
1
NIL
HORIZONTAL

@#$#@#$#@
## WHAT IS IT?

A pair of artifical neural networks that implement the language mapping model described in Couch (2016). The main use of this model is to serve as the child model for ConversationWorld.nlogo, though it can be explored on its own terms.

## HOW IT WORKS

The model consists of two multi-layer artifical neural networks. The two ANNs are yolked togather, such that the reverse

## HOW TO USE IT



## THINGS TO NOTICE


## THINGS TO TRY


## RELATED MODELS

(models in the NetLogo Models Library and elsewhere which are of related interest)

## CREDITS AND REFERENCES

Big props to the models of Artificial Neural Networks contained in the modelling library. I would not have been able to do this without them.
@#$#@#$#@
default
true
0
Polygon -7500403 true true 150 5 40 250 150 205 260 250

airplane
true
0
Polygon -7500403 true true 150 0 135 15 120 60 120 105 15 165 15 195 120 180 135 240 105 270 120 285 150 270 180 285 210 270 165 240 180 180 285 195 285 165 180 105 180 60 165 15

arrow
true
0
Polygon -7500403 true true 150 0 0 150 105 150 105 293 195 293 195 150 300 150

box
false
0
Polygon -7500403 true true 150 285 285 225 285 75 150 135
Polygon -7500403 true true 150 135 15 75 150 15 285 75
Polygon -7500403 true true 15 75 15 225 150 285 150 135
Line -16777216 false 150 285 150 135
Line -16777216 false 150 135 15 75
Line -16777216 false 150 135 285 75

bug
true
0
Circle -7500403 true true 96 182 108
Circle -7500403 true true 110 127 80
Circle -7500403 true true 110 75 80
Line -7500403 true 150 100 80 30
Line -7500403 true 150 100 220 30

butterfly
true
0
Polygon -7500403 true true 150 165 209 199 225 225 225 255 195 270 165 255 150 240
Polygon -7500403 true true 150 165 89 198 75 225 75 255 105 270 135 255 150 240
Polygon -7500403 true true 139 148 100 105 55 90 25 90 10 105 10 135 25 180 40 195 85 194 139 163
Polygon -7500403 true true 162 150 200 105 245 90 275 90 290 105 290 135 275 180 260 195 215 195 162 165
Polygon -16777216 true false 150 255 135 225 120 150 135 120 150 105 165 120 180 150 165 225
Circle -16777216 true false 135 90 30
Line -16777216 false 150 105 195 60
Line -16777216 false 150 105 105 60

car
false
0
Polygon -7500403 true true 300 180 279 164 261 144 240 135 226 132 213 106 203 84 185 63 159 50 135 50 75 60 0 150 0 165 0 225 300 225 300 180
Circle -16777216 true false 180 180 90
Circle -16777216 true false 30 180 90
Polygon -16777216 true false 162 80 132 78 134 135 209 135 194 105 189 96 180 89
Circle -7500403 true true 47 195 58
Circle -7500403 true true 195 195 58

circle
false
0
Circle -7500403 true true 0 0 300

circle 2
false
0
Circle -7500403 true true 0 0 300
Circle -16777216 true false 30 30 240

cow
false
0
Polygon -7500403 true true 200 193 197 249 179 249 177 196 166 187 140 189 93 191 78 179 72 211 49 209 48 181 37 149 25 120 25 89 45 72 103 84 179 75 198 76 252 64 272 81 293 103 285 121 255 121 242 118 224 167
Polygon -7500403 true true 73 210 86 251 62 249 48 208
Polygon -7500403 true true 25 114 16 195 9 204 23 213 25 200 39 123

cylinder
false
0
Circle -7500403 true true 0 0 300

dot
false
0
Circle -7500403 true true 90 90 120

face happy
false
0
Circle -7500403 true true 8 8 285
Circle -16777216 true false 60 75 60
Circle -16777216 true false 180 75 60
Polygon -16777216 true false 150 255 90 239 62 213 47 191 67 179 90 203 109 218 150 225 192 218 210 203 227 181 251 194 236 217 212 240

face neutral
false
0
Circle -7500403 true true 8 7 285
Circle -16777216 true false 60 75 60
Circle -16777216 true false 180 75 60
Rectangle -16777216 true false 60 195 240 225

face sad
false
0
Circle -7500403 true true 8 8 285
Circle -16777216 true false 60 75 60
Circle -16777216 true false 180 75 60
Polygon -16777216 true false 150 168 90 184 62 210 47 232 67 244 90 220 109 205 150 198 192 205 210 220 227 242 251 229 236 206 212 183

fish
false
0
Polygon -1 true false 44 131 21 87 15 86 0 120 15 150 0 180 13 214 20 212 45 166
Polygon -1 true false 135 195 119 235 95 218 76 210 46 204 60 165
Polygon -1 true false 75 45 83 77 71 103 86 114 166 78 135 60
Polygon -7500403 true true 30 136 151 77 226 81 280 119 292 146 292 160 287 170 270 195 195 210 151 212 30 166
Circle -16777216 true false 215 106 30

flag
false
0
Rectangle -7500403 true true 60 15 75 300
Polygon -7500403 true true 90 150 270 90 90 30
Line -7500403 true 75 135 90 135
Line -7500403 true 75 45 90 45

flower
false
0
Polygon -10899396 true false 135 120 165 165 180 210 180 240 150 300 165 300 195 240 195 195 165 135
Circle -7500403 true true 85 132 38
Circle -7500403 true true 130 147 38
Circle -7500403 true true 192 85 38
Circle -7500403 true true 85 40 38
Circle -7500403 true true 177 40 38
Circle -7500403 true true 177 132 38
Circle -7500403 true true 70 85 38
Circle -7500403 true true 130 25 38
Circle -7500403 true true 96 51 108
Circle -16777216 true false 113 68 74
Polygon -10899396 true false 189 233 219 188 249 173 279 188 234 218
Polygon -10899396 true false 180 255 150 210 105 210 75 240 135 240

house
false
0
Rectangle -7500403 true true 45 120 255 285
Rectangle -16777216 true false 120 210 180 285
Polygon -7500403 true true 15 120 150 15 285 120
Line -16777216 false 30 120 270 120

leaf
false
0
Polygon -7500403 true true 150 210 135 195 120 210 60 210 30 195 60 180 60 165 15 135 30 120 15 105 40 104 45 90 60 90 90 105 105 120 120 120 105 60 120 60 135 30 150 15 165 30 180 60 195 60 180 120 195 120 210 105 240 90 255 90 263 104 285 105 270 120 285 135 240 165 240 180 270 195 240 210 180 210 165 195
Polygon -7500403 true true 135 195 135 240 120 255 105 255 105 285 135 285 165 240 165 195

line
true
0
Line -7500403 true 150 0 150 300

line half
true
0
Line -7500403 true 150 0 150 150

pentagon
false
0
Polygon -7500403 true true 150 15 15 120 60 285 240 285 285 120

person
false
0
Circle -7500403 true true 110 5 80
Polygon -7500403 true true 105 90 120 195 90 285 105 300 135 300 150 225 165 300 195 300 210 285 180 195 195 90
Rectangle -7500403 true true 127 79 172 94
Polygon -7500403 true true 195 90 240 150 225 180 165 105
Polygon -7500403 true true 105 90 60 150 75 180 135 105

plant
false
0
Rectangle -7500403 true true 135 90 165 300
Polygon -7500403 true true 135 255 90 210 45 195 75 255 135 285
Polygon -7500403 true true 165 255 210 210 255 195 225 255 165 285
Polygon -7500403 true true 135 180 90 135 45 120 75 180 135 210
Polygon -7500403 true true 165 180 165 210 225 180 255 120 210 135
Polygon -7500403 true true 135 105 90 60 45 45 75 105 135 135
Polygon -7500403 true true 165 105 165 135 225 105 255 45 210 60
Polygon -7500403 true true 135 90 120 45 150 15 180 45 165 90

sheep
false
15
Circle -1 true true 203 65 88
Circle -1 true true 70 65 162
Circle -1 true true 150 105 120
Polygon -7500403 true false 218 120 240 165 255 165 278 120
Circle -7500403 true false 214 72 67
Rectangle -1 true true 164 223 179 298
Polygon -1 true true 45 285 30 285 30 240 15 195 45 210
Circle -1 true true 3 83 150
Rectangle -1 true true 65 221 80 296
Polygon -1 true true 195 285 210 285 210 240 240 210 195 210
Polygon -7500403 true false 276 85 285 105 302 99 294 83
Polygon -7500403 true false 219 85 210 105 193 99 201 83

square
false
0
Rectangle -7500403 true true 30 30 270 270

square 2
false
0
Rectangle -7500403 true true 30 30 270 270
Rectangle -16777216 true false 60 60 240 240

star
false
0
Polygon -7500403 true true 151 1 185 108 298 108 207 175 242 282 151 216 59 282 94 175 3 108 116 108

target
false
0
Circle -7500403 true true 0 0 300
Circle -16777216 true false 30 30 240
Circle -7500403 true true 60 60 180
Circle -16777216 true false 90 90 120
Circle -7500403 true true 120 120 60

tree
false
0
Circle -7500403 true true 118 3 94
Rectangle -6459832 true false 120 195 180 300
Circle -7500403 true true 65 21 108
Circle -7500403 true true 116 41 127
Circle -7500403 true true 45 90 120
Circle -7500403 true true 104 74 152

triangle
false
0
Polygon -7500403 true true 150 30 15 255 285 255

triangle 2
false
0
Polygon -7500403 true true 150 30 15 255 285 255
Polygon -16777216 true false 151 99 225 223 75 224

truck
false
0
Rectangle -7500403 true true 4 45 195 187
Polygon -7500403 true true 296 193 296 150 259 134 244 104 208 104 207 194
Rectangle -1 true false 195 60 195 105
Polygon -16777216 true false 238 112 252 141 219 141 218 112
Circle -16777216 true false 234 174 42
Rectangle -7500403 true true 181 185 214 194
Circle -16777216 true false 144 174 42
Circle -16777216 true false 24 174 42
Circle -7500403 false true 24 174 42
Circle -7500403 false true 144 174 42
Circle -7500403 false true 234 174 42

turtle
true
0
Polygon -10899396 true false 215 204 240 233 246 254 228 266 215 252 193 210
Polygon -10899396 true false 195 90 225 75 245 75 260 89 269 108 261 124 240 105 225 105 210 105
Polygon -10899396 true false 105 90 75 75 55 75 40 89 31 108 39 124 60 105 75 105 90 105
Polygon -10899396 true false 132 85 134 64 107 51 108 17 150 2 192 18 192 52 169 65 172 87
Polygon -10899396 true false 85 204 60 233 54 254 72 266 85 252 107 210
Polygon -7500403 true true 119 75 179 75 209 101 224 135 220 225 175 261 128 261 81 224 74 135 88 99

wheel
false
0
Circle -7500403 true true 3 3 294
Circle -16777216 true false 30 30 240
Line -7500403 true 150 285 150 15
Line -7500403 true 15 150 285 150
Circle -7500403 true true 120 120 60
Line -7500403 true 216 40 79 269
Line -7500403 true 40 84 269 221
Line -7500403 true 40 216 269 79
Line -7500403 true 84 40 221 269

wolf
false
0
Polygon -16777216 true false 253 133 245 131 245 133
Polygon -7500403 true true 2 194 13 197 30 191 38 193 38 205 20 226 20 257 27 265 38 266 40 260 31 253 31 230 60 206 68 198 75 209 66 228 65 243 82 261 84 268 100 267 103 261 77 239 79 231 100 207 98 196 119 201 143 202 160 195 166 210 172 213 173 238 167 251 160 248 154 265 169 264 178 247 186 240 198 260 200 271 217 271 219 262 207 258 195 230 192 198 210 184 227 164 242 144 259 145 284 151 277 141 293 140 299 134 297 127 273 119 270 105
Polygon -7500403 true true -1 195 14 180 36 166 40 153 53 140 82 131 134 133 159 126 188 115 227 108 236 102 238 98 268 86 269 92 281 87 269 103 269 113

x
false
0
Polygon -7500403 true true 270 75 225 30 30 225 75 270
Polygon -7500403 true true 30 75 75 30 270 225 225 270

@#$#@#$#@
NetLogo 6.0-M6
@#$#@#$#@
@#$#@#$#@
@#$#@#$#@
@#$#@#$#@
@#$#@#$#@
default
0.0
-0.2 0 0.0 1.0
0.0 1 1.0 0.0
0.2 0 0.0 1.0
link direction
true
0
Line -7500403 true 150 150 90 180
Line -7500403 true 150 150 210 180

@#$#@#$#@
0
@#$#@#$#@
